import unittest

from packages.manifest_utils import get_nodes_and_dependencies

class TestParseManifest(unittest.TestCase):
    def test_parse_manifest(self):
        manifest = {
            'sources': {
                'source1': {},
                'source2': {}
            },
            'nodes': {
                'node1': {
                    'resource_type': 'model',
                    'depends_on': {
                        'nodes': ['source1', 'source2']
                    }
                },
                'node2': {
                    'resource_type': 'seed',
                    'depends_on': {
                        'nodes': ['node1']
                    }
                }
            }
        }
        expected_nodes = ['source1', 'source2', 'node1', 'node2']
        expected_dependencies = {
            'node1': ['source1', 'source2'],
            'node2': ['node1'],
            'source1': [],
            'source2': []
        }

        nodes, dependencies = get_nodes_and_dependencies(manifest=manifest)

        self.assertListEqual(nodes, expected_nodes)
        self.assertDictEqual(dependencies, expected_dependencies)

    def test_parse_manifest_with_default_seed(self):
        manifest = {
            'sources': {
                'source1': {},
                'source2': {}
            },
            'nodes': {
                'model1': {
                    'resource_type': 'model',
                    'depends_on': {
                        'nodes': ['source1', 'source2']
                    }
                },
                'model2': {
                    'resource_type': 'model',
                    'depends_on': {
                        'nodes': ['model1']
                    }
                },
                'seed1': {
                    'resource_type': 'seed',
                    'depends_on': {
                        'nodes': ['model2']
                    }
                }
            }
        }
        default_seed = 'default_seed'
        expected_dependencies = {
            'model1': ['source1', 'source2'],
            'model2': ['model1'],
            'seed1': ['model2'],
            'source1': [],
            'source2': [],
        }
        nodes, dependencies = get_nodes_and_dependencies(manifest, default_seed)
        self.assertDictEqual(dependencies, expected_dependencies)

    # Test matching nodes from the manifest based on regex
    def test_parse_manifest_with_node_match_regex(self):
        manifest = {
            'sources': {
                'source1': {},
                'source2': {}
            },
            'nodes': {
                'model.xyz1': {
                    'resource_type': 'model',
                    'depends_on': {
                        'nodes': ['source1', 'source2']
                    }
                },
                'model.xyz2': {
                    'resource_type': 'model',
                    'depends_on': {
                        'nodes': ['model.xyz1']
                    }
                },
                'model.not_a_matching_model': {
                    'resource_type': 'model',
                    'depends_on': {
                        'nodes': ['model.xyz1']
                    }
                },             
                'test.not_a_matching_model': {
                    'resource_type': 'model',
                    'depends_on': {
                        'nodes': ['model.xyz1']
                    }
                },                       
                'seed1': {
                    'resource_type': 'seed',
                    'depends_on': {
                        'nodes': []
                    }
                }
            }
        }
        node_match_regex = 'xyz*'
        expected_dependencies = {
            'model.xyz1': ['source1', 'source2'],
            'model.xyz2': ['model.xyz1'],
            'seed1': [],
            'source1': [],
            'source2': [],
        }
        _, dependencies = get_nodes_and_dependencies(manifest, node_match_regex=node_match_regex)
        self.assertDictEqual(dependencies, expected_dependencies)


if __name__ == '__main__':
    unittest.main()


import unittest

from packages.manifest_utils import get_tests_and_nodes

class TestGetTestsAndNodes(unittest.TestCase):
    # Returns a tuple with two dictionaries.
    def test_returns_tuple_with_two_dictionaries(self):
        # Arrange
        manifest = {
            'nodes': {
                'node1': {
                    'resource_type': 'test',
                    'name': 'test1',
                    'depends_on': {
                        'nodes': ['node2']
                    }
                },
                'node2': {
                    'resource_type': 'model',
                    'name': 'model1'
                }
            }
        }
    
        # Act
        result = get_tests_and_nodes(manifest)
    
        # Assert
        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], dict)
        self.assertIsInstance(result[1], dict)

    # The first dictionary contains the tests and their data.
    def test_first_dictionary_contains_tests_and_data(self):
        # Arrange
        manifest = {
            'nodes': {
                'node1': {
                    'resource_type': 'test',
                    'name': 'test1',
                    'depends_on': {
                        'nodes': ['node2']
                    }
                },
                'node2': {
                    'resource_type': 'model',
                    'name': 'model1'
                }
            }
        }
    
        # Act
        result = get_tests_and_nodes(manifest)
    
        # Assert
        self.assertIn('test1', result[0])
        self.assertEqual(result[0]['test1'], manifest['nodes']['node1'])

    # The second dictionary contains the nodes and the tests that depend on them.
    def test_second_dictionary_contains_nodes_and_tests(self):
        # Arrange
        manifest = {
            'nodes': {
                'node1': {
                    'resource_type': 'test',
                    'name': 'test1',
                    'depends_on': {
                        'nodes': ['node2']
                    }
                },
                'node2': {
                    'resource_type': 'model',
                    'name': 'model1'
                }
            }
        }
    
        # Act
        result = get_tests_and_nodes(manifest)
    
        # Assert
        self.assertIn('node2', result[1])
        self.assertEqual(result[1]['node2'], ['test1'])


import unittest

from packages.manifest_utils import sort_topologically

class TestTopologicalSort(unittest.TestCase):
    def test_simple_graph(self):
        graph = {'a': ['b', 'c'], 'b': ['c'], 'c': []}
        expected_order = ['c', 'b', 'a']
        self.assertEqual(sort_topologically(graph), expected_order)

    def test_graph_with_cycle(self):
        graph = {'a': ['b'], 'b': ['c'], 'c': ['a']}
        with self.assertRaises(ValueError):
            sort_topologically(graph)

    def test_disconnected_graph(self):
        graph = {'a': ['b'], 'b': ['c'], 'c': [], 'd': []}
        expected_order = ['c', 'b', 'a', 'd']
        self.assertEqual(sort_topologically(graph), expected_order)


import json
import click
import re
from typing import Tuple

def get_tests_and_nodes(manifest :dict) -> Tuple[dict, dict]:
    """
    Get the tests and nodes that the tests depend on.
    """
    tests_with_data = {}
    nodes_with_tests = {}
    for _, node_data in manifest.get('nodes', {}).items():
        if node_data.get('resource_type') == 'test':
            tests_with_data[node_data.get('name')] = node_data
            
            # Get the models that the test depends on
            for node in node_data.get('depends_on', {}).get('nodes', []):
                nodes_with_tests.setdefault(node, []).append(node_data.get('name'))

    return tests_with_data, nodes_with_tests

def get_ascend_node_name(node_name :str) -> str:
    """
    Get the Ascend node name from the dbt node name.
    """
    return node_name.split('.')[-1].lower()

def get_dbt_node_type(node_name :str) -> str:
    """
    Get the dbt node type from the dbt node name.
    """
    return node_name.split('.')[0]

def get_dbt_node_name(nodes :list, node :str) -> str:
    """
    Get the dbt node name from the dbt node name.
    """
    result = [node_str for node_str in nodes if node_str.split('.')[-1] == node]

    if len(result) == 0:
        raise ValueError(f"Node {node} not found in nodes list")

    return result[0]

def get_nodes_and_dependencies(manifest :dict, default_seed :str = None, node_match_regex :str = "*") -> Tuple[list, dict[str, list[str]]]:    
    """
    Parse the manifest files and return a list of nodes in topological order and a dict of dependencies of each node.
    """
    # Extract source elements
    sources = []
    for node_name, _ in manifest['sources'].items():
        sources.append(node_name)

    # Extract model and seed dependencies
    dependencies = {}
    for node_name, node_data in manifest['nodes'].items():
        if (node_data['resource_type'] == 'model' and node_matches("model.", node_name, node_match_regex)) or node_data['resource_type'] == 'seed':
            # Check if 'depends_on' and 'nodes' keys exist
            if 'depends_on' in node_data and 'nodes' in node_data['depends_on']:
                dependencies[node_name] = node_data['depends_on']['nodes']
            else:
                dependencies[node_name] = []

    # Perform topological sort
    sorted_nodes = sort_topologically(dependencies)

    # Replace empty dependencies with default seed unless it is a source node
    for node in sorted_nodes:
        # If dependency is a source, add empty list
        if node in sources:
            dependencies[node] = []
        # If dependency is empty and node is a model, add default seed
        elif len(dependencies[node]) == 0 and node.split('.')[0] == 'model':
            dependencies[node] = [default_seed] if default_seed else []

    # return list of sorted nodes and dict of dependencies of each node
    return sorted_nodes, dependencies


def sort_topologically(graph):
    """
    Sort nodes in topological order.
    """
    visited = set()
    post_order = []
    temp_marked = set()

    def visit(node):
        if node in temp_marked:
            raise ValueError("Graph contains a cycle")
        if node not in visited:
            temp_marked.add(node)
            for neighbor in graph.get(node, []):
                visit(neighbor)
            temp_marked.remove(node)
            visited.add(node)
            post_order.append(node)

    for node in graph:
        visit(node)

    return post_order

@click.pass_context
def get_compiled_sql(ctx, node_str):
    """
    Get compiled SQL from manifest.
    """
    return ctx.obj['manifest']['nodes'][node_str]['compiled_code']

@click.pass_context
def get_description(ctx, node_str):
    """
    Get description from manifest.
    """
    return ctx.obj['manifest']['nodes'][node_str].get('description', '')

def node_matches(prefix :str, node_name :str, node_match_regex :str) -> bool:
    """
    Check if the node name matches the node match regex.
    """
    return node_match_regex == '*' or re.match(prefix + node_match_regex, node_name) is not None

@click.pass_context
def get_reduction_from_manifest(ctx, node_str :str) -> str:
    """
    Get reduction from manifest JSON. Values are ['no-reduction', 'full-reduction']. The default is 'no-reduction'
        Will be added to dbt_project.yml as a custom config parameter:
        ```
        models:
            jaffle_shop:
                weather-models-dbt:
                    stg_weather_data:
                        ascend-partition-reduction: full-reduction
        ```
    """
    return ctx.obj['manifest']['nodes'][node_str].get('config', {}).get('ascend-partition-reduction', 'no-reduction')


import unittest

from packages import transform_utils as tu

class TestTranslateSql(unittest.TestCase):

    def test_simple_snowflake_sql(self):
        sql = "SELECT * FROM XX-XX.YYY.my_table"
        input_ids = ['my_table']

        result = tu.translate_sql("snowflake-sql", sql=sql, input_ids=input_ids)

        expected_sql = "SELECT * FROM {{my_table}}"
        self.assertEqual(result, expected_sql)

    def test_more_complex_snowflake_sql(self):
        sql = "SELECT * FROM XX-XX.YYY.table1, XX-XX.YYY.table2"
        input_ids = ['table2', 'table1']

        result = tu.translate_sql("snowflake-sql", sql=sql, input_ids=input_ids)

        expected_sql = "SELECT * FROM {{table1}}, {{table2}}"
        self.assertEqual(result, expected_sql)

    def test_simple_bigquery_sql(self):
        sql = "SELECT * FROM XX-XX.YYY.my_table"
        input_ids = ['my_table']

        result = tu.translate_sql("bigquery-sql", sql=sql, input_ids=input_ids)

        expected_sql = "SELECT * FROM {{my_table}}"
        self.assertEqual(result, expected_sql)

    def test_quoted_bigquery_sql(self):
        sql = "SELECT * FROM `XX-XX`.`YYY`.`my_table`"
        input_ids = ['my_table']

        result = tu.translate_sql("bigquery-sql", sql=sql, input_ids=input_ids)

        expected_sql = "SELECT * FROM {{my_table}}"
        self.assertEqual(result, expected_sql)

    def test_simple_databricks_sql(self):
        sql = "SELECT * FROM `XXX`.`YYY`.`my_table`"
        input_ids = ['my_table']

        result = tu.translate_sql("databricks-sql", sql=sql, input_ids=input_ids)

        expected_sql = "SELECT * FROM {{my_table}}"
        self.assertEqual(result, expected_sql)

    def test_wrong_databricks_sql(self):
        sql = "SELECT * FROM `XXX`.`YYY`.`ZZZ`"
        input_ids = ['my_table']

        result = tu.translate_sql("databricks-sql", sql=sql, input_ids=input_ids)

        self.assertEqual(result, sql)

    def test_wrong_bq_sql(self):
        sql = "SELECT * FROM XXX.YYY.ZZZ"
        input_ids = ['my_table']

        result = tu.translate_sql("bigquery-sql", sql=sql, input_ids=input_ids)

        self.assertEqual(result, sql)

    def test_mesh_snowflake_sql_1(self):

        self.maxDiff = None

        sql = """
                with

                customers as (

                    select * from ASCEND__MHYATT_SF_DEMO.None.stg_customers

                ),

                orders_table as (

                    select * from ASCEND__MHYATT_SF_DEMO.None.orders

                ),

                order_items_table as (

                    select * from ASCEND__MHYATT_SF_DEMO.None.order_items
                ),

                order_summary as (

                    select
                        orders.customer_id,

                        count(distinct orders.order_id) as count_lifetime_orders,
                        count(distinct orders.order_id) > 1 as is_repeat_buyer,
                        min(orders.ordered_at) as first_ordered_at,
                        max(orders.ordered_at) as last_ordered_at,
                        sum(order_items.product_price) as lifetime_spend_pretax,
                        sum(orders.order_total) as lifetime_spend

                    from orders_table as orders

                    left join
                        order_items_table as order_items
                        on orders.order_id = order_items.order_id

                    group by 1

                ),

                joined as (

                    select
                        customers.*,
                        order_summary.count_lifetime_orders,
                        order_summary.first_ordered_at,
                        order_summary.last_ordered_at,
                        order_summary.lifetime_spend_pretax,
                        order_summary.lifetime_spend,

                        case
                            when order_summary.is_repeat_buyer then 'returning'
                            else 'new'
                        end as customer_type

                    from customers

                    left join order_summary
                        on customers.customer_id = order_summary.customer_id

                )
        """.strip()
        sql_out = """
                with

                customers as (

                    select * from {{stg_customers}}

                ),

                orders_table as (

                    select * from {{orders}}

                ),

                order_items_table as (

                    select * from {{order_items}}
                ),

                order_summary as (

                    select
                        orders.customer_id,

                        count(distinct orders.order_id) as count_lifetime_orders,
                        count(distinct orders.order_id) > 1 as is_repeat_buyer,
                        min(orders.ordered_at) as first_ordered_at,
                        max(orders.ordered_at) as last_ordered_at,
                        sum(order_items.product_price) as lifetime_spend_pretax,
                        sum(orders.order_total) as lifetime_spend

                    from orders_table as orders

                    left join
                        order_items_table as order_items
                        on orders.order_id = order_items.order_id

                    group by 1

                ),

                joined as (

                    select
                        customers.*,
                        order_summary.count_lifetime_orders,
                        order_summary.first_ordered_at,
                        order_summary.last_ordered_at,
                        order_summary.lifetime_spend_pretax,
                        order_summary.lifetime_spend,

                        case
                            when order_summary.is_repeat_buyer then 'returning'
                            else 'new'
                        end as customer_type

                    from customers

                    left join order_summary
                        on customers.customer_id = order_summary.customer_id

                )
        """.strip()

        input_ids = ['stg_customers', 'orders', 'order_items']

        result = tu.translate_sql("snowflake-sql", sql=sql, input_ids=input_ids)

        self.assertEqual(result, sql_out)        

    def test_mesh_snowflake_sql_2(self):

        self.maxDiff = None

        sql = """
                with

                orders as (

                    select * from ASCEND__MHYATT_SF_DEMO.None.stg_orders

                ),

                order_items_table as (

                    select * from ASCEND__MHYATT_SF_DEMO.PUBLIC.order_items

                ),

                order_items_summary as (

                    select

                        order_id,
                        count(case when is_food_item then 1 else 0 end)
                            as count_food_items,
                        count(case when is_drink_item then 1 else 0 end)
                            as count_drink_items,

                        sum(supply_cost) as order_cost

                    from order_items_table

                    group by 1

                ),


                compute_booleans as (
                    select

                        orders.*,
                        order_items_summary.order_cost,
                        order_items_summary.count_food_items > 0 as is_food_order,
                        order_items_summary.count_drink_items > 0 as is_drink_order

                    from orders

                    left join
                        order_items_summary
                        on orders.order_id = order_items_summary.order_id
                )

                select * from compute_booleans
        """.strip()
        sql_out = """
                with

                orders as (

                    select * from {{stg_orders}}

                ),

                order_items_table as (

                    select * from {{order_items}}

                ),

                order_items_summary as (

                    select

                        order_id,
                        count(case when is_food_item then 1 else 0 end)
                            as count_food_items,
                        count(case when is_drink_item then 1 else 0 end)
                            as count_drink_items,

                        sum(supply_cost) as order_cost

                    from order_items_table

                    group by 1

                ),


                compute_booleans as (
                    select

                        orders.*,
                        order_items_summary.order_cost,
                        order_items_summary.count_food_items > 0 as is_food_order,
                        order_items_summary.count_drink_items > 0 as is_drink_order

                    from orders

                    left join
                        order_items_summary
                        on orders.order_id = order_items_summary.order_id
                )

                select * from compute_booleans
        """.strip()

        input_ids = ['stg_orders', 'order_items']

        result = tu.translate_sql("snowflake-sql", sql=sql, input_ids=input_ids)

        self.assertEqual(result, sql_out)   
     


import click
import json

from dotenv import load_dotenv

from packages import process_module
from packages import run_tests_module
from packages import component_module

# Silently try to load the .env file if it exists
load_dotenv()

# Top-level group
@click.group(
        help="""
        Collection of utilities to help convert dbt projects into Ascend dataflows.
        """)
def cli():
    pass

# sub-group defining operations to perform on a collection of dbt models together
@cli.group(
        help="""
        Operations to perform on a collection of dbt models together.
        """)
def dbt_models():
    pass

# sub-group defining operations to perform on dbt tests
@cli.group(
        help="""
        Operations to perform on dbt tests.
        """)
def dbt_tests():
    pass

# sub-group defining operations to perform on an individual dbt model
@cli.group(
        help="""
        Operations to perform on an individual dbt model.
        """)
def dbt_model():
    pass

def ascend_options(f):
    f = click.option('--hostname', required=True, help='Ascend hostname. envvar=ASCEND_HOSTNAME', envvar='ASCEND_HOSTNAME')(f)
    f = click.option('--data-service', required=True, help='Ascend data service name, envvar=ASCEND_DATA_SERVICE', envvar='ASCEND_DATA_SERVICE')(f)
    f = click.option('--dataflow', required=True, help='Ascend dataflow name, envvar=ASCEND_DATAFLOW', envvar='ASCEND_DATAFLOW')(f)
    f = click.option('--language', required=False, help='Language to use for dbt transform translation, envvar=ASCEND_LANGUAGE"', default="snowflake-sql", envvar='ASCEND_LANGUAGE', type=click.Choice(['snowflake-sql', 'bigquery-sql', 'databricks-sql']))(f)
    return f

def load_json(ctx, param, value):
    """
    Callback to load the JSON file into the context object when --manifest-file is passed.
    """
    if not value or ctx.resilient_parsing:
        return
    try:
        with open(value, 'r') as f:
            manifest = json.load(f)
        ctx.ensure_object(dict)
        ctx.obj['manifest'] = manifest
    except Exception as e:
        raise click.BadParameter(f"Could not load JSON file: {e}")
    return value

def manifest_options(f):
    f = click.option('--manifest-file', required=True, help='Path to the manifest JSON file, envvar=DBT_MANIFEST_FILE', envvar='DBT_MANIFEST_FILE', type=click.Path(exists=True), callback=load_json)(f)
    f = click.option('--model-selector', required=False, help='Regular expression to only process the matching models, envvar=DBT_MODEL_MATCHING_REGEX', envvar='DBT_MODEL_MATCHING_REGEX', default="*")(f)
    return f


@dbt_models.command()
@click.pass_context
@ascend_options
@manifest_options
@click.option('--default-seed', required=False, help='Default seed to connect hanging models to. Defaults to one of the nodes in the dataflow, envvar=DEFAULT_SEED', envvar='DEFAULT_SEED')
def merge(ctx, **kwargs):
    """Process the compiled dbt manifest and SQL files and create/update/delete Ascend dataflow transforms."""
    process_module.merge_cmd(ctx, **kwargs)
dbt_models.add_command(merge)

@dbt_models.command()
@click.pass_context
@ascend_options
@manifest_options
def update_sql(ctx, **kwargs):
    """Update the SQL of existing Ascend dataflow transforms."""
    process_module.update_sql_cmd(ctx, **kwargs)
dbt_models.add_command(update_sql)

@dbt_models.command()
@click.pass_context
@ascend_options
@manifest_options
def delete(ctx, **kwargs):
    """Delete all dbt models from an Ascend dataflow."""
    process_module.delete_cmd(ctx, **kwargs)
dbt_models.add_command(delete)

@dbt_models.command()
@click.pass_context
@ascend_options
@manifest_options
def validate(ctx, **kwargs):
    """Validate the seeds and sources are present in the dataflow."""
    process_module.validate_cmd(ctx, **kwargs)
dbt_models.add_command(validate)

@dbt_models.command()
@click.pass_context
@ascend_options
@manifest_options
def verify(ctx, **kwargs):
    """Verify the seeds and sources are present in the dataflow."""
    process_module.validate_cmd(ctx, **kwargs)
dbt_models.add_command(validate)

@dbt_models.command()
@click.pass_context
@manifest_options
def show(ctx, **kwargs):
    """Show the dependencies of dbt models."""
    process_module.show_cmd(ctx, **kwargs)
dbt_models.add_command(show)

@dbt_model.command()
@click.pass_context
@ascend_options
@manifest_options
@click.option('--model-name', required=True, help='Name of the dbt model to create.')
@click.option('--reduction', required=False, help='Partition reduction to use for the component. Defaults to "no-reduction.', default='no-reduction', type=click.Choice(['no-reduction', 'full-reduction']))
@click.option('--with-tests', required=False, help='Create tests for the component.', default=False, is_flag=True)
def create_component(ctx, **kwargs):
    """Create a component from a dbt model."""
    component_module.create_component_cmd(ctx, **kwargs)
dbt_model.add_command(create_component)

@dbt_model.command()
@click.pass_context
@ascend_options
@manifest_options
@click.option('--model-name', required=True, help='Name of the dbt model to update with the new SQL from the manifest.')
def update_component_sql(ctx, **kwargs):
    """Update the SQL of an existing component."""
    component_module.update_component_sql_cmd(ctx, **kwargs)
dbt_model.add_command(update_component_sql)

@dbt_model.command()
@click.pass_context
@ascend_options
@manifest_options
@click.option('--model-name', required=True, help='Name of the dbt model to show the SQL from the manifest.')
@click.option('--show-raw', required=False, help='Show the raw SQL from the manifest.', default=False, is_flag=True)
def show_component_sql(ctx, **kwargs):
    """Show the SQL of an existing component."""
    component_module.show_component_sql_cmd(ctx, **kwargs)
dbt_model.add_command(show_component_sql)

@dbt_tests.command()
@click.pass_context
@ascend_options
@manifest_options
def deploy_tests(ctx, **kwargs):
    """Create and run all dbt tests against the deployed models in Ascend."""
    run_tests_module.deploy_tests_cmd(ctx, **kwargs)
dbt_tests.add_command(deploy_tests)

@dbt_tests.command()
@click.pass_context
@ascend_options
@manifest_options
def delete_tests(ctx, **kwargs):
    """Delete all dbt tests against the deployed models in Ascend."""
    run_tests_module.delete_tests_cmd(ctx, **kwargs)
dbt_tests.add_command(delete_tests)

@dbt_tests.command()
@click.pass_context
@ascend_options
@manifest_options
def check_test_results(ctx, **kwargs):
    """Check the results of all dbt tests against the deployed models in Ascend."""
    run_tests_module.check_test_results_cmd(ctx, **kwargs)
dbt_tests.add_command(check_test_results)


if __name__ == "__main__":
    cli(prog_name='ascend_dbt_utils')


from ascend.sdk import definitions
from ascend.sdk.applier import DataflowApplier
from ascend.sdk.client import Client

from . import manifest_utils as mu
from . import transform_utils as tu


# This is currently only handling models, sources and seeds
def show_cmd(ctx, **kwargs):
    """
    Load and parse manifest JSON file and print the dependencies of dbt models.
    """

    nodes, dependencies = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], node_match_regex=ctx.params['model_selector'])

    # Print nodes in topological order
    for node in nodes:
        print(f"\nNode: {node}. Depends on:")
        for dependency in dependencies[node]:
            print(f"  -> {dependency}")

def merge_cmd(ctx, **kwargs):
    """
    Merge dbt models into an Ascend dataflow.
    """
    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    # Get existing nodes in dataflow
    existing_nodes = client.list_dataflow_components(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], deep=True).data
    existing_node_ids = [node.id for node in existing_nodes]

    nodes, dependencies = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], default_seed=existing_node_ids[0] if ctx.params['default_seed'] is None else ctx.params['default_seed'], node_match_regex=ctx.params['model_selector'])

    # Ensure seed nodes present in the manifest are present in the dataflow. Exit if not.
    for node_str in nodes:
        if mu.get_dbt_node_type(node_str) in ['seed', 'source']:
            node = mu.get_ascend_node_name(node_str)
            if node not in existing_node_ids:
                print(f"Seed node {node} is not present in the dataflow. Please add it manually and try again.")
                exit(1)

    # For every "model" node create a component
    components = []
    for node_str in nodes:

        # Skip if node is not a model
        if mu.get_dbt_node_type(node_str) != 'model':
            continue

        node = mu.get_ascend_node_name(node_str)
        # Create component
        component = tu.create_transform(
            id=node, 
            sql=mu.get_compiled_sql(node_str=node_str), 
            inputs=dependencies[node_str],
            description=mu.get_description(node_str=node_str),
            language=ctx.params['language'],
            reduction=mu.get_reduction_from_manifest(node_str=node_str),
        )
        # Add component to list of existing nodes
        components.append(component)

    # Get dataflow definition
    dataflow_def = client.get_dataflow(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow']).data

    # Perform a non-deleting append
    applier = DataflowApplier(client)
    applier.apply(data_service_id=ctx.params['data_service'], dataflow=definitions.Dataflow(id=dataflow_def.id, name=dataflow_def.name, components=components), delete=False, dry_run=False)

def delete_cmd(ctx, **kwargs):
    """
    Delete dbt models from Ascend dataflow.
    """
    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    nodes, _ = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], node_match_regex=ctx.params['model_selector'])

    # Remove all nodes from the list of existing nodes in reverse order
    node_ids = [mu.get_ascend_node_name(node) for node in reversed(nodes) if mu.get_dbt_node_type(node) == 'model']
    for node in node_ids:
        print(f"Deleting transform {node}")
        try:
            client.delete_transform(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], id=node)
        except Exception as e:
            print(f"Could not delete transform {node}. Error: {e.reason}")

def validate_cmd(ctx, **kwargs):
    """
    Validate the seeds are present in the dataflow.
    """
    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    nodes, _ = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], node_match_regex=ctx.params['model_selector'])

    # Validate the seeds are present in the dataflow
    node_ids = [mu.get_ascend_node_name(node) for node in nodes if mu.get_dbt_node_type(node) in ['seed', 'source']]

    # Get existing nodes in dataflow
    existing_nodes = client.list_dataflow_components(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], deep=True).data
    existing_node_ids = [node.id for node in existing_nodes]

    # Print the list of nodes present and absent in the dataflow
    print("Nodes present in the dataflow:")
    for node in node_ids:
        if node in existing_node_ids:
            print(f"  {node}")
    print("Nodes absent in the dataflow:")
    for node in node_ids:
        if node not in existing_node_ids:
            print(f"  {node}")


def update_sql_cmd(ctx, **kwargs):
    """
    Update the SQL of existing Ascend dataflow transforms.
    """

    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    nodes, dependencies = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], node_match_regex=ctx.params['model_selector'])

    # For every node, if there is a transform with the same name, update the SQL. Otherwise, display a message.
    for node_str in nodes:
        node_type = mu.get_dbt_node_type(node_str)

        if node_type != 'model':
            continue

        node = mu.get_ascend_node_name(node_str)
        input_ids = [mu.get_ascend_node_name(node) for node in dependencies[node_str]]
        print(f"Updating SQL of the transform {node}")
        try:
            transform_body = client.get_transform(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], id=node).data
            transform_body.view.operator.spark_function.executable.code.source.inline = tu.translate_sql(
                language=ctx.params['language'],
                sql=mu.get_compiled_sql(node_str=node_str), 
                input_ids=input_ids)
            client.update_transform(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], transform_id=node, body=transform_body)
        except Exception as e:
            print(f"Could not update transform {node}. Error: {e.reason}") 





import ascend.protos.component.component_pb2 as component
import ascend.protos.function.function_pb2 as function
import ascend.protos.io.io_pb2 as io
import ascend.protos.operator.operator_pb2 as operator

from ascend.sdk import definitions

import re
import click

def create_transform(
    id,
    sql,
    inputs,
    description,
    custom_dq_test_sql=None,
    reduction="no-reduction",
    language="snowflake-sql",
):
    """
    Create Ascend transform component from dbt model.
    """
    input_ids = [x.split(".")[-1] for x in inputs]

    # In sql, find all the occurrences of the input nodes with DB name and schema name using regex as per the following example
    # Example input: select * from XXX.YYY.customer_revenue_by_month
    # Example output: select * from {{ customer_revenue_by_month }}
    prev_sql = sql
    sql = translate_sql(language=language, sql=sql, input_ids=input_ids)

    # If no substitutions were made, display the error and exit
    # Ascend does not allow a transform to have no inputs
    if sql == prev_sql:
        print(
            f"No inputs replaced in the file {id}.sql. Please check the file and try again."
        )
        exit(1)

    # Create transform component
    # TODO: Add support for other SQL flavours
    # TODO: Add support for other partition reduction options
    return definitions.Transform(
        id=id,
        name=id,
        description=description,
        input_ids=input_ids,
        operator=operator.Operator(
            spark_function=operator.Spark.Function(
                executable=io.Executable(
                    code=io.Code(
                        language=get_language(language),
                        source=io.Code.Source(
                            inline=sql,
                        ),
                    ),
                ),
                reduction=get_reduction(reduction_method=reduction),
                tests=define_quality_tests(custom_dq_test_sql),
            ),
        ),
        assigned_priority=component.Priority(),
    )


def get_language(type):
    """
    Get the language for the transform component.
    """

    if type == "snowflake-sql":
        return function.Code.Language(
            snowflake_sql=function.Code.Language.SnowflakeSql()
        )
    
    if type == "bigquery-sql":
        return function.Code.Language(
            bigquery_sql=function.Code.Language.BigQuerySql()
        )

    if type == "databricks-sql":
        return function.Code.Language(
            databricks_sql=function.Code.Language.DatabricksSql()
        )

    raise Exception(f"Language type {type} not supported.")


def get_reduction(reduction_method):
    """
    Get the reduction method for the transform component.
    """

    if reduction_method == "no-reduction":
        return operator.Reduction(
            no_reduction=operator.Reduction.NoReduction(),
        )
    elif reduction_method == "full-reduction":
        return operator.Reduction(full=operator.Reduction.Full())
    else:
        raise Exception(f"Reduction method {reduction_method} not supported.")

def translate_sql(language, sql, input_ids):
    """
    Translate the SQL to replace the input nodes with Ascend placeholders.
    """
    for input in input_ids:

        # Fix to remove all "`" characters from 3 part names
        sql = re.sub(r"`([\w\-]+)`\.`([\w\-]+)`\.`({})`".format(input), r"\1.\2.\3", sql, flags=re.IGNORECASE)

        sql = re.sub(r"[\w\-]+\.[\w\-]+\.({})".format(input), lambda m: "{{{{{}}}}}".format(m.group(1).lower()), sql, flags=re.IGNORECASE)
        
    return sql.strip()


# TODO: Add support for standard checks
def define_quality_tests(custom_dq_test_sql):
    """
    Define quality tests for a transform component. Only supports no checks and custom checks.
    """

    if custom_dq_test_sql is None:
        return function.QualityTests()

    return function.QualityTests(
        custom=[
            function.QualityTests.CustomCheck(
                sql=custom_dq_test_sql.strip(),
            )
        ]
    )



from setuptools import setup, find_packages

setup(
    name='ascend-io-dbt-utils',
    version='0.11.3',
    description='Utilities for dbt and Ascend',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(
        exclude=['tests', 'tests.*'],
        ),
    entry_points='''
        [console_scripts]
        ascend_dbt_utils=packages.cli:cli
    ''',
    install_requires=[
        'ascend-io-sdk==0.2.62',
        'click==8.1.7',
        'python-dotenv==1.0.0',
    ],
    url='https://github.com/michaelhyatt/ascend_dbt_transform'
)


from ascend.sdk import definitions
from ascend.sdk.applier import DataflowApplier
from ascend.sdk.client import Client

import json

from . import manifest_utils as mu
from . import transform_utils as tu

def update_component_sql_cmd(ctx, **kwargs):
    """
    Update the SQL of existing Ascend dataflow transform.
    """
    node = ctx.params['model_name']
    print(f"Updating SQL of the transform {node}")

    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    nodes, dependencies = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], node_match_regex=ctx.params['model_selector'])

    node_str = mu.get_dbt_node_name(nodes, node)

    input_ids = [mu.get_ascend_node_name(node) for node in dependencies[node_str]]

    try:
        transform_body = client.get_transform(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], id=node).data
        transform_body.view.operator.spark_function.executable.code.source.inline = tu.translate_sql(
            language=ctx.params['language'],
            sql=mu.get_compiled_sql(node_str=node_str), 
            input_ids=input_ids)
        client.update_transform(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], transform_id=node, body=transform_body)
    except Exception as e:
        print(f"Could not update transform {node}. Error: {e.reason}")
        if e.body is not None:
            error_body = json.loads(e.body)
            if error_body.get('errors') is not None:
                print(f"Returned from API error: {error_body['errors']}")


def create_component_cmd(ctx, **kwargs):
    """
    Create a component from a dbt model.
    """
    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    # Get manifest nodes and their dependencies
    nodes, dependencies = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], node_match_regex=ctx.params['model_selector'])

    # Get the list of existing dataflow components
    existing_nodes = client.list_dataflow_components(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], deep=True).data

    # Check that the component has all its dependencies in the dataflow
    model_name = ctx.params['model_name']

    # Check that the model_name exists
    try:
        mu.get_dbt_node_name(nodes, model_name)
    except:
        print(f"Node {model_name} is not present in the dataflow. Please check it and try again.")
        exit(1)

    node_id = mu.get_dbt_node_name(nodes, model_name)
    for node_str in dependencies[node_id]:
        node = mu.get_ascend_node_name(node_str)
        if  node not in [node.id for node in existing_nodes]:
            print(f"Node {node} is not present in the dataflow. Please add it and try again.")
            exit(1)

    # Create the component
    component = tu.create_transform(
        id=ctx.params['model_name'], 
        sql=mu.get_compiled_sql(node_str=node_id), 
        inputs=[mu.get_ascend_node_name(node) for node in dependencies[node_id]],
        description=mu.get_description(node_str=node_id),
        reduction=ctx.params['reduction'],
        language=ctx.params['language'],
    )
    components = [component]

    # Deploy tests if requested
    groups = []
    if ctx.params['with_tests']:

        # Get test data
        test_data, nodes_with_test_data = mu.get_tests_and_nodes(manifest=ctx.obj['manifest'])

        # Get tests for the created node
        test_id = [test for test in nodes_with_test_data if mu.get_ascend_node_name(test) == ctx.params['model_name']][0]
        tests = nodes_with_test_data[test_id] 

        # Create transforms for the tests
        for test_id in tests:
            test = test_data[test_id]

            # If one of the inputs is not there, don't deploy the test and report it
            test_dependencies = test['depends_on']['nodes']
            existing_components_plus_transform = {node.id for node in existing_nodes}
            existing_components_plus_transform.add(ctx.params['model_name'])
            missing_nodes = {mu.get_ascend_node_name(dep) for dep in test_dependencies} - existing_components_plus_transform
            if len(missing_nodes) > 0:
                print(f"Test {test['name']} is missing the following dependencies: {missing_nodes}. Skipping test.")
                continue

            components.append(
                tu.create_transform(
                    id=test['name'], 
                    sql=test['compiled_code'], 
                    inputs=test['depends_on']['nodes'],
                    description=test.get('description', ''),
                    custom_dq_test_sql='select * from {{target}} WHERE 1=1',
                    reduction='full-reduction',
                    language=ctx.params['language'],
                )
            )

        groups = [definitions.ComponentGroup(
            id=f'{ctx.params["model_name"]}_dbt_tests',
            name=f'{ctx.params["model_name"]}_dbt_tests',
            component_ids = [transform.id for transform in components if transform.id != ctx.params['model_name']],
            description=f'{ctx.params["model_name"]} dbt tests'
            )]

    # Get dataflow definition
    dataflow_def = client.get_dataflow(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow']).data

    # Perform a non-deleting append
    applier = DataflowApplier(client)
    applier.apply(data_service_id=ctx.params['data_service'], dataflow=definitions.Dataflow(id=dataflow_def.id, name=dataflow_def.name, components=components, groups=groups), delete=False, dry_run=False)

def show_component_sql_cmd(ctx, **kwargs):
    """
    Show the SQL of existing Ascend dataflow transform.
    """

    nodes, dependencies = mu.get_nodes_and_dependencies(manifest=ctx.obj['manifest'], node_match_regex=ctx.params['model_selector'])
    node = mu.get_dbt_node_name(nodes, ctx.params['model_name'])
    print(f"Showing SQL of the {node}\n ------------------\n")

    compiled_sql = mu.get_compiled_sql(node_str=node)

    if ctx.params['show_raw']:
        print(compiled_sql)
    else:
        print(tu.translate_sql(
            language=ctx.params['language'],
            sql=compiled_sql, 
            input_ids=[mu.get_ascend_node_name(node) for node in dependencies[node]])
            )


from ascend.sdk import definitions
from ascend.sdk.applier import DataflowApplier
from ascend.sdk.client import Client

from . import manifest_utils as mu
from . import transform_utils as tu


def deploy_tests_cmd(ctx, **kwargs):

    # Get nodes and dependencies
    test_data, nodes_with_test_data = mu.get_tests_and_nodes(manifest=ctx.obj['manifest'])

    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    # Get the list of components in the dataflow
    components = client.list_dataflow_components(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], deep=True).data

    # For every node with test data, check the node exists in the dataflow
    nodes_to_be_tested = []
    for node_to_be_tested in [node for node in nodes_with_test_data]:
        node_id = mu.get_ascend_node_name(node_to_be_tested)
        if node_id not in [component.id for component in components]:
            print(f"Node {node_id} is not present in the dataflow.")
        else:
            print(f"Node {node_id} is present in the dataflow.")
            nodes_to_be_tested.append(node_to_be_tested)

    # For every existing node from the list, add transforms for every test linked to that node
    transforms = []
    present_component_names = [component.id for component in components]
    for test, test_details in test_data.items():

        # If one of the inputs is not there, don't deploy the test and report it
        test_dependencies = [mu.get_ascend_node_name(test) for test in test_details['depends_on']['nodes']]
        missing_nodes = set(test_dependencies) - set(present_component_names)
        if len(missing_nodes) > 0:
            print(f"Test {test} is missing the following dependencies: {missing_nodes}. Skipping test.")
            continue

        # If the test has no compiled code, skip the test and report an error
        if not 'compiled_code' in test_details.keys():
            print(f"Test {test} has no compiled code. Skipping test.")
            continue

        print(f"Adding test {test_details['name']} to the dataflow.")

        transforms.append(
            tu.create_transform(
                id=test_details['name'], 
                sql=test_details['compiled_code'], 
                inputs=test_details['depends_on']['nodes'],
                description=test_details.get('description', ''),
                custom_dq_test_sql='select * from {{target}} WHERE 1=1',
                reduction='full-reduction',
                language=ctx.params['language'],
            )
        )

    # Check if there are any transforms to deploy
    if len(transforms) == 0:
        print("No tests to deploy.")
        return

    # Create a group and add all the transforms to the group
    group=definitions.ComponentGroup(
        id="dbt_tests",
        name="dbt tests",
        component_ids = [transform.id for transform in transforms],
        description="dbt tests"
        )

    # Deploy the test transforms using the dataflow applier
    # Get dataflow definition
    dataflow_def = client.get_dataflow(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow']).data

    # Create dataflow definition
    dataflow_def = definitions.Dataflow(
        id=dataflow_def.id, 
        name=dataflow_def.name, 
        components=transforms,
        groups=[group]
        )

    # Perform a non-deleting append
    applier = DataflowApplier(client)
    applier.apply(data_service_id=ctx.params['data_service'], dataflow=dataflow_def, delete=False, dry_run=False)


def delete_tests_cmd(ctx, **kwargs):
    # Create Ascend Client
    client = Client(ctx.params['hostname'])
    
    # Get test nodes
    test_data, _ = mu.get_tests_and_nodes(manifest=ctx.obj['manifest'])

    # Remove all nodes from the list of existing nodes in reverse order
    node_ids = [node for node in test_data.keys()]
    for node in node_ids:
        print(f"Deleting test {node}")
        try:
            client.delete_transform(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], id=node)
        except Exception as e:
            print(f"An error occurred while deleting transform {node}: {e.reason}")

    # Delete component group
    print(f"Deleting component group dbt_tests")
    try:
        client.delete_component_group(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], id="dbt_tests")
    except Exception as e:
        print(f"An error occurred while deleting dbt_tests group: {e.reason}")

def check_test_results_cmd(ctx, **kwargs):
    """
    Check the results of all dbt tests against the deployed models in Ascend.
    """
    # Create Ascend Client
    client = Client(ctx.params['hostname'])

    # Get test nodes
    test_data, _ = mu.get_tests_and_nodes(manifest=ctx.obj['manifest'])

    # Get the list of components in the dataflow
    components = client.list_dataflow_components(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], deep=True).data

    # For every node with test data, check the node exists in the dataflow
    nodes_to_be_tested = []
    for node_to_be_tested in [node for node in test_data.keys()]:
        node_id = mu.get_ascend_node_name(node_to_be_tested)
        if node_id not in [component.id for component in components]:
            print(f"Node {node_id} is not present in the dataflow.")
        else:
            nodes_to_be_tested.append(node_id)

    # For every existing test node from the list, retrieve and print the number of records
    print("Checking test results based on the number of records returned by tests...")
    for test_node in nodes_to_be_tested:
        try:
            test_results = client.get_transform_records(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], id=test_node, limit=1).data
            num_of_records = len(test_results.data.records_response.records.value)
            message = f"{Colours.GREEN}Ok.{Colours.RESET}" if num_of_records == 0 else f"{Colours.RED}Failed.{Colours.RESET}"
            print(f"Test {test_node} produced {num_of_records} records. Test result is {message}")
        except Exception as e:
            print(f"An error occurred while retrieving test {test_node} results: {e.reason}")

    # For every existing test node from the list, check custom SQL test results
    print()
    print("Checking test results based on data quality check results...")
    for test_node in nodes_to_be_tested:
        try:
            test_results = client.get_transform_quality_test_results(data_service_id=ctx.params['data_service'], dataflow_id=ctx.params['dataflow'], id=test_node).data
            result = test_results.test_summaries[0].result
            if result == "passed":
                result_str = f"{Colours.GREEN}{result}{Colours.RESET}"
            else:
                result_str = f"{Colours.RED}{result}{Colours.RESET}"
            print(f"Test {test_node} produced {result_str}")
        except Exception as e:
            print(f"An error occurred while retrieving test {test_node} results: {e.reason}")

class Colours:
    RED = '\033[31m'
    GREEN = '\033[32m'
    RESET = '\033[0m'